from ewah.uploaders.base import EWAHBaseUploader
from ewah.hooks.postgres import EWAHPostgresHook
from ewah.constants import EWAHConstants as EC
from ewah.utils.airflow_utils import PGO

from psycopg2.extras import execute_values
from copy import deepcopy

import hashlib
import re


class EWAHPostgresUploader(EWAHBaseUploader):

    _QUERY_SCHEMA_CHANGES_COLUMNS = """
        SELECT
        	f.attname AS "name"
        FROM pg_attribute f
        	JOIN pg_class cl ON cl.OID = f.attrelid
        	LEFT JOIN pg_namespace n ON n.OID = cl.relnamespace
        WHERE cl.relkind = 'r'::CHAR
        	AND n.nspname = %(schema_name)s
        	AND cl.relname = %(table_name)s
        	AND f.attnum > 0;
    """
    _QUERY_SCHEMA_CHANGES_ADD_COLUMN = """
        ALTER TABLE "{schema_name}"."{table_name}"
        ADD COLUMN "{column_name}" {column_type};
    """
    _QUERY_TABLE = 'SELECT * FROM "{schema_name}"."{table_name}"'

    _COPY_TABLE = """
        -- Drop a previous version of the table if it exists
        DROP TABLE IF EXISTS "{new_schema}"."{new_table}";
        CREATE TABLE "{new_schema}"."{new_table}" AS (
            SELECT * FROM "{old_schema}"."{old_table}"
        );
    """

    CONSTRAINTS_SET = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(EC.DWH_ENGINE_POSTGRES, *args, **kwargs)

    @classmethod
    def get_schema_tasks(
        cls,
        dag,
        dwh_engine,
        dwh_conn_id,
        target_schema_name,
        target_schema_suffix="_next",
        target_database_name=None,
        read_right_users=None,  # Only for PostgreSQL
        **additional_task_args,
    ):
        sql_kickoff = """
            DROP SCHEMA IF EXISTS "{schema_name}{schema_suffix}" CASCADE;
            CREATE SCHEMA "{schema_name}{schema_suffix}";
        """.format(
            schema_name=target_schema_name,
            schema_suffix=target_schema_suffix,
        )
        sql_final = """
            DROP SCHEMA IF EXISTS "{schema_name}" CASCADE;
            ALTER SCHEMA "{schema_name}{schema_suffix}"
                RENAME TO "{schema_name}";
        """.format(
            schema_name=target_schema_name,
            schema_suffix=target_schema_suffix,
        )

        # Don't fail final task just because a user or role that should
        # be granted read rights does not exist!
        grant_rights_sql = """
            DO $$
            BEGIN
              GRANT USAGE ON SCHEMA "{target_schema_name}" TO {user};
              GRANT SELECT ON ALL TABLES
                IN SCHEMA "{target_schema_name}" TO {user};
              EXCEPTION WHEN OTHERS THEN -- catches any error
                RAISE NOTICE 'not granting rights - user does not exist!';
            END
            $$;
        """
        if read_right_users:
            if not isinstance(read_right_users, list):
                raise Exception("Arg read_right_users must be of type List!")
            for user in read_right_users:
                if re.search(r"\s", user) or (";" in user):
                    _msg = "No whitespace or semicolons allowed in usernames!"
                    raise ValueError(_msg)
                sql_final += grant_rights_sql.format(
                    target_schema_name=target_schema_name,
                    user=user,
                )

        task_1_args = deepcopy(additional_task_args)
        task_2_args = deepcopy(additional_task_args)
        task_1_args.update(
            {
                "sql": sql_kickoff,
                "task_id": "kickoff",
                "dag": dag,
                "postgres_conn_id": dwh_conn_id,
            }
        )
        task_2_args.update(
            {
                "sql": sql_final,
                "task_id": "final",
                "dag": dag,
                "postgres_conn_id": dwh_conn_id,
            }
        )
        return (PGO(**task_1_args), PGO(**task_2_args))

    def commit(self):
        self.dwh_hook.commit()

    def rollback(self):
        self.dwh_hook.commit()

    def close(self):
        self.dwh_hook.close()

    def _create_or_update_table(
        self,
        data,
        table_name,
        schema_name,
        schema_suffix,
        columns_definition,
        load_strategy,
        upload_call_count,
        primary_key=None,
    ):
        self.log.info("Preparing DWH Tables...")
        schema_name += schema_suffix
        if (upload_call_count == 1 and load_strategy == EC.LS_INSERT_REPLACE) or (
            not self.test_if_table_exists(
                table_name=table_name,
                schema_name=schema_name,
            )
        ):
            self.dwh_hook.execute(
                sql="""
                    DROP TABLE IF EXISTS "{schema_name}"."{table_name}" CASCADE;
                    CREATE TABLE "{schema_name}"."{table_name}" ({columns});
                """.format(
                    schema_name=schema_name,
                    table_name=table_name,
                    columns=",\n\t".join(
                        [
                            '"{0}"\t{1}'.format(col, self._get_column_type(defi))
                            for col, defi in columns_definition.items()
                        ]
                    ),
                ),
                commit=False,
            )
            if primary_key:
                self.dwh_hook.execute(
                    sql="""
                        ALTER TABLE ONLY "{schema_name}"."{table_name}"
                        ADD PRIMARY KEY ("{columns}");
                    """.format(
                        schema_name=schema_name,
                        table_name=table_name,
                        columns='","'.join(primary_key),
                    )
                )

        if primary_key and not self.CONSTRAINTS_SET:
            # make sure there is a unique constraint for primary_key
            self.CONSTRAINTS_SET = True  # Only set it once per DagRun, though
            self.dwh_hook.execute(
                sql="""
                    ALTER TABLE "{schema_name}"."{table_name}"
                    DROP CONSTRAINT IF EXISTS "{constraint}";
                    ALTER TABLE "{schema_name}"."{table_name}"
                    ADD CONSTRAINT "{constraint}" UNIQUE ("{columns}");
                """.format(
                    schema_name=schema_name,
                    table_name=table_name,
                    # max length for constraint is 63 chars - make sure it's unique!
                    constraint="__ewah_{0}".format(
                        hashlib.blake2b(
                            "ufu_{0}_{1}".format(schema_name, table_name).encode(),
                            digest_size=28,
                        ).hexdigest()
                    ),
                    columns='", "'.join(primary_key),
                ),
                commit=False,
            )

        set_columns = []
        for column in columns_definition.keys():
            if not (column in (primary_key or [])):
                set_columns += [column]

        if load_strategy == EC.LS_INSERT_REPLACE:
            # Ought not be a conflict - just insert
            do_on_conflict = ""
        elif load_strategy == EC.LS_INSERT_ADD:
            # Ought not be a conflict - just insert
            do_on_conflict = ""
        elif load_strategy == EC.LS_UPSERT:
            # Update records as appropriate
            do_on_conflict = """
                ON CONFLICT ("{primary_key}") DO UPDATE SET\n\t{sets}
            """.format(
                primary_key='", "'.join(primary_key),
                sets="\n\t,".join(
                    [
                        '"{column}" = EXCLUDED."{column}"'.format(column=column)
                        for column in set_columns
                    ]
                ),
            )
        else:
            raise Exception("Not implemented!")

        cols_list = list(columns_definition.keys())
        sql = (
            """
            INSERT INTO "{schema_name}"."{table_name}"
            ("{column_names}") VALUES {placeholder}
            {do_on_conflict};
        """.format(
                schema_name=schema_name,
                table_name=table_name,
                placeholder="{placeholder}",
                column_names='", "'.join(cols_list),
                do_on_conflict=do_on_conflict,
            )
            .replace("%", "%%")
            .format(placeholder="%s")
        )
        self.log.info("Now Uploading! Using SQL:\n\n{0}".format(sql))
        # escape crappy column names by using aliases in the psycopg2 template
        cols_map = {cols_list[i]: "col_" + str(i) for i in range(len(cols_list))}
        template = "(%(" + ")s, %(".join([val for key, val in cols_map.items()]) + ")s)"
        cur = self.dwh_hook.cursor
        upload_data = [
            {cols_map[key]: row.get(key) for key in cols_list} for row in data
        ]
        execute_values(
            cur=cur,
            sql=sql,
            argslist=upload_data,
            template=template,
        )
        self.log.info("Upload done.")

        self.dwh_hook.execute(
            sql='ANALYZE "{0}"."{1}";'.format(schema_name, table_name),
            commit=False,
        )

    def test_if_table_exists(self, table_name, schema_name):
        return bool(
            self.dwh_hook.execute_and_return_result(
                sql="SELECT to_regclass(%(name)s);",
                params={"name": '"{0}"."{1}"'.format(schema_name, table_name)},
            )[0][0]
        )

    def get_max_value_of_column(self, column_name):
        return self.dwh_hook.execute_and_return_result(
            sql='SELECT MAX("{0}") FROM {1}'.format(
                column_name,
                f'"{self.schema_name}{self.schema_suffix}"."{self.table_name}"',
            ),
            return_dict=False,
        )[0][0]



from datetime import datetime
from yahoofinancials import YahooFinancials

from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator
from ewah.utils.airflow_utils import datetime_utcnow_with_tz


class EWAHFXOperator(EWAHBaseOperator):

    _NAMES = ["fx"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
    }

    def __init__(
        self,
        currency_pair,  # iterable of length 2
        frequency="daily",  # daily, weekly, or monthly
        *args,
        **kwargs
    ):

        if not frequency in ("daily", "weekly", "monthly"):
            raise Exception("Frequency must be one of: daily, weekly, monthly")

        if not len(currency_pair) == 2:
            raise Exception(
                "currency_pair must be iterable of length 2 "
                + "containing the currency pair."
            )

        # No connection id required for this operator
        if kwargs.get("source_conn_id"):
            self.log.info(
                "source_conn_id is not required for operator! Ignoring argument."
            )
        kwargs["source_conn_id"] = None
        # If incremental, the primary key column is 'date'
        if kwargs.get("primary_key"):
            self.log.info("primary_key is fixed for this operator. Using the default.")
        kwargs["primary_key"] = ["date"]

        self.currency_pair = currency_pair
        self.frequency = frequency

        super().__init__(*args, **kwargs)

    def ewah_execute(self, context):
        data_from = self.data_from or context["dag"].start_date
        data_until = self.data_until or datetime_utcnow_with_tz()

        format_str = "%Y-%m-%d"
        currency_str = "{0}{1}=X".format(*self.currency_pair)
        data = YahooFinancials([currency_str]).get_historical_price_data(
            data_from.strftime(format_str),
            data_until.strftime(format_str),
            self.frequency,
        )
        self.upload_data(data[currency_str]["prices"])



from ewah.hooks.base import EWAHBaseHook

from typing import List, Dict, Any, Optional

import requests


class EWAHRapidmailHook(EWAHBaseHook):

    _ATTR_RELABEL: {}

    conn_name_attr = "ewah_rapidmail_conn_id"
    default_conn_name = "ewah_rapidmail_default"
    conn_type = "ewah_rapidmail"
    hook_name = "EWAH Rapidmail Connection"

    URL = "https://apiv3.emailsys.net"
    ALLOWED_ENDPOINTS = (
        "apiusers",
        "blacklist",
        "forms",
        "mailings",
        "recipientlists",
        "recipients",
        "trx/emails",
    )

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra", "host", "port", "schema"],
            "relabeling": {
                "login": "API User ID",
                "password": "API User Password",
            },
        }

    def get_all_data(self, endpoint: str) -> List[Dict[str, Any]]:
        assert endpoint in self.ALLOWED_ENDPOINTS, "Invalid endpoint '{0}'!".format(
            endpoint
        )
        # special case: recipients can only be fetched by recipientlist_id
        if endpoint == "recipients":
            recipient_lists = self.get_data(endpoint="recipientlists")
            for recipient_list in recipient_lists:
                params = {"recipientlist_id": recipient_list["id"]}
                yield self.get_data(endpoint=endpoint, params=params)
        else:
            yield self.get_data(endpoint=endpoint)

    def get_data(
        self, endpoint: str, params: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Any]]:
        data = []
        page = 1
        params = params or {}

        auth = requests.auth.HTTPBasicAuth(self.conn.login, self.conn.password)

        while True:
            self.log.info("Requesting page {0}...".format(page))
            params["page"] = page
            request = requests.get(
                "/".join((self.URL, endpoint)), auth=auth, params=params
            )
            assert request.status_code == 200, request.text
            response = request.json()
            data += response["_embedded"][endpoint.replace("/", "")]
            if page == 1:
                self.log.info(
                    "Total pages: {0}".format(
                        response["page_count"]
                        or 1  # If no data exists, page_count is 0
                    )
                )
            if response["page"] == response["page_count"]:
                # we're done here
                break
            page += 1

        return data



from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator
from ewah.hooks.amazon_seller_central import EWAHAmazonSellerCentralHook

from datetime import datetime, timedelta

import pendulum
import time
import pytz


class EWAHAmazonSellerCentralReportsAPIOperator(EWAHBaseOperator):
    """To be used with the Reporting API"""

    _NAMES = ["seller_central_reporting", "amazon_reporting"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        # Depends on individual report, dict is updated during __init__
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: False,
    }

    def __init__(
        self,
        marketplace_region,
        report_name,
        report_options=None,
        ewah_options=None,
        *args,
        **kwargs,
    ):

        assert EWAHAmazonSellerCentralHook.validate_marketplace_region(
            marketplace_region, allow_lists=True
        ), f"Marketplace Region {marketplace_region} is invalid!"
        assert (
            report_name in EWAHAmazonSellerCentralHook._REPORT_METADATA.keys()
        ), "Invalid report name!"
        report_metadata = EWAHAmazonSellerCentralHook._REPORT_METADATA[report_name]
        kwargs["primary_key"] = report_metadata["primary_key"]
        kwargs["subsequent_field"] = report_metadata["subsequent_field"]

        # Update accepted extract strategies based on individual report
        for strategy in report_metadata.get("accepted_strategies", []):
            self._ACCEPTED_EXTRACT_STRATEGIES[strategy] = True

        super().__init__(*args, **kwargs)

        self.marketplace_region = marketplace_region
        self.report_name = report_name
        self.report_options = report_options
        self.ewah_options = ewah_options

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field) - (
                self.load_data_from_relative or timedelta(days=0)
            )
            data_until = self.data_until or datetime.utcnow().replace(tzinfo=pytz.utc)
        else:
            data_from = self.data_from
            data_until = self.data_until or datetime.utcnow().replace(tzinfo=pytz.utc)

        for batch in self.source_hook.get_data_from_reporting_api_in_batches(
            marketplace_region=self.marketplace_region,
            report_name=self.report_name,
            data_from=data_from,
            data_until=data_until,
            report_options=self.report_options,
            ewah_options=self.ewah_options,
        ):
            self.upload_data(batch)



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.linkedin import EWAHLinkedInHook

from datetime import datetime, timedelta, date


class EWAHLinkedInOperator(EWAHBaseOperator):

    _NAMES = ["li", "linkedin"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: False,
    }

    _CONN_TYPE = EWAHLinkedInHook.conn_type

    def __init__(self, object_name=None, *args, **kwargs):
        object_name = object_name or kwargs["target_table_name"]
        assert (
            object_name in EWAHLinkedInHook._AVAILABLE_OBJECTS
        ), "Invalid object {0}!".format(object_name)
        super().__init__(*args, **kwargs)

        self.object_name = object_name

    def ewah_execute(self, context):
        self.upload_data(self.source_hook.get_object_data(self.object_name))



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.linkedin import EWAHLinkedInHook

from datetime import datetime, timedelta, date


class EWAHLinkedInAdsOperator(EWAHBaseOperator):

    _NAMES = ["li_ads", "linkedin_ads"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHLinkedInHook.conn_type

    def __init__(self, pivot, fields, account_ids=None, *args, **kwargs):
        fields = list(set(fields + list(EWAHLinkedInHook._REQUIRED_FIELDS_ANALYTICS)))
        pivot = pivot.upper()
        EWAHLinkedInHook.validate_fields_list(fields)
        kwargs["primary_key"] = ["date_start", "pivotValue"]
        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = "date_start"
        super().__init__(*args, **kwargs)

        if isinstance(account_ids, str):
            account_ids = [account_ids]

        if not account_ids is None:
            assert isinstance(
                account_ids, list
            ), "account_ids must be a list of strings!"

        self.pivot = pivot
        self.fields = fields
        self.account_ids = account_ids

    def ewah_execute(self, context):
        date_end = (self.data_until or datetime.now()).date()
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            date_start = self.get_max_value_of_column(self.subsequent_field) - (
                self.load_data_from_relative or timedelta(days=0)
            )
        else:
            date_start = self.data_from.date()

        if self.load_data_chunking_timedelta:
            batch_from = date_start
            while batch_from <= date_end:
                batch_until = min(
                    date_end, batch_from + self.load_data_chunking_timedelta
                )
                for batch in self.source_hook.get_analytics_in_batches(
                    pivot=self.pivot,
                    date_start=batch_from,
                    date_end=batch_until,
                    fields=self.fields,
                    account_ids=self.account_ids,
                ):
                    self.upload_data(batch)
                batch_from = batch_until + timedelta(days=1)
        else:
            for batch in self.source_hook.get_analytics_in_batches(
                pivot=self.pivot,
                date_start=date_start,
                date_end=date_end,
                fields=self.fields,
                account_ids=self.account_ids,
            ):
                self.upload_data(batch)



from ewah.hooks.base import EWAHBaseHook

from typing import Optional, Dict, Any, Union, List


class EWAHSQLBaseHook(EWAHBaseHook):
    """Base hook extension for use as parent of various SQL hooks.

    Children need the following:
    - _get_db_conn method
    - _get_cursor property
    - _get_dictcursor property
    - execute method
    - get_data_from_sql method
    """

    _DEFAULT_PORT = 1234  # overwrite in child

    @property
    def dbconn(self):
        if not hasattr(self, "_dbconn"):
            if hasattr(self.conn, "ssh_conn_id") and self.conn.ssh_conn_id:
                if not hasattr(self, "_ssh_hook"):
                    self._ssh_hook = EWAHBaseHook.get_hook_from_conn_id(
                        conn_id=self.conn.ssh_conn_id
                    )
                    self.local_bind_address = self._ssh_hook.start_tunnel(
                        self.conn.host, self.conn.port or self._DEFAULT_PORT
                    )
            else:
                self.local_bind_address = self.conn.host, self.conn.port
            self._dbconn = self._get_db_conn()
        return self._dbconn

    @property
    def cursor(self):
        """Cursor that returns lists of lists from the data source."""
        if not hasattr(self, "_cur"):
            self._cur = self._get_cursor()
        return self._cur

    @property
    def dictcursor(self):
        """Cursor that returns lists of dictionaries from the data source."""
        if not hasattr(self, "_dictcur"):
            self._dictcur = self._get_dictcursor()
        return self._dictcur

    def get_records(self, sql, parameters=None):
        """
        Variant of execute method. Required to work with the SQL sensor.
        """
        return self.execute_and_return_result(
            sql=sql, params=parameters, return_dict=False
        )

    def execute_and_return_result(
        self,
        sql: str,
        params: Optional[Dict[str, Any]] = None,
        return_dict: bool = False,
    ) -> Union[List[list], List[dict]]:
        cursor = self.dictcursor if return_dict else self.cursor
        self.execute(sql=sql, params=params, commit=False, cursor=cursor)
        return cursor.fetchall()

    def get_data_in_batches(
        self,
        sql: str,
        params: Optional[dict] = None,
        return_dict: bool = True,
        batch_size: int = 10000,
    ):
        cur = self.dictcursor if return_dict else self.cursor
        self.execute(sql, params=params, cursor=cur, commit=False)
        while True:
            self.log.info("Fetching next batch...")
            data = cur.fetchmany(batch_size)
            if data:
                yield data
            else:
                break

    def commit(self):
        self.log.info("Committing changes!")
        return self.dbconn.commit()

    def rollback(self):
        self.log.info("Rolling back changes!")
        return self.dbconn.rollback()

    def close(self):
        if hasattr(self, "_cur"):
            self._cur.close()
            del self._cur
        if hasattr(self, "_dictcur"):
            self._dictcur.close()
            del self._dictcur
        if hasattr(self, "_dbconn"):
            if hasattr(self, "_ssh_hook"):
                self._ssh_hook.stop_tunnel()
                del self._ssh_hook
            self._dbconn.close()
            del self._dbconn

    def __del__(self):
        self.close()



from ewah.hooks.base import EWAHBaseHook

import requests


class EWAHSevDeskHook(EWAHBaseHook):

    conn_name_attr = "ewah_sevdesk_conn_id"
    default_conn_name = "ewah_sevdesk_default"
    conn_type = "ewah_sevdesk"
    hook_name = "EWAH SevDesk Connection"

    _ENDPOINTS = [
        "CheckAccount",
        "CheckAccountTransaction",
        "CommunicationWay",
        "Contact",
        "Invoice",
        "InvoicePos",
        "Order",
        "OrderPos",
        "Part",
        "Voucher",
        "VoucherPos",
    ]
    _BASE_URL = "https://my.sevdesk.de/api/v1/"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "schema", "extra", "host", "login"],
            "relabeling": {
                "password": "API Token",
            },
        }

    @classmethod
    def validate_endpoint(cls, endpoint):
        return endpoint in cls._ENDPOINTS

    def get_data_in_batches(self, endpoint, batch_size=10000, embed=None):
        assert self.validate_endpoint(endpoint), "Invalid endpoint '{0}'!".format(
            endpoint
        )
        url = self._BASE_URL + endpoint
        params = {
            "limit": batch_size,
            "offset": 0,
            "token": self.conn.password,
        }
        if embed:
            if not isinstance(embed, str):
                # Cast iterable into CSV
                embed = ",".join(embed)
            params["embed"] = embed
        self.log.info("Request URL: {0}".format(url))
        while True:
            self.log.info("Requesting a new page...")
            request = requests.get(url, params=params)
            assert request.status_code == 200, request.text
            response = request.json()["objects"]
            if response:
                params["offset"] = params["offset"] + len(response)
                yield response
            else:
                # We're done
                break
